/*
Copyright (C) 2003-2007 Anders Ervik <calvin@countzero.no>
Copyright (C) 2006-2007 Leonard Ritter

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <memory.h>
#include <RtAudio.h>
#include <assert.h>
#include "driver.h"

namespace zzub {

audiodriver::audiodriver()
{
	desired_framerate = 48000;
	desired_framecount = 256;
	audio = 0;
	started = false;
    recording = false;
    defaultDevice = -1;
}

void audiodriver::setSampleRate(unsigned int samplerate) {
	desired_framerate = samplerate;
}

unsigned int audiodriver::getSampleRate() {
	return desired_framerate;
}

void audiodriver::setBufferSize(unsigned int buffersize) {
	desired_framecount = buffersize;
}

unsigned int audiodriver::getBufferSize() {
	return desired_framecount;
}

void i2s(float **s, float *i, int numsamples) {
	if (!numsamples)
		return;
	float *p[] = {s[0],s[1]};
	while (numsamples--)
	{
		*p[0]++ = *i++;
		*p[1]++ = *i++;
	};
}

int rtaudio_process_callback(char *output, int frameCount, void *data) {
	assert(frameCount <= audiodriver::MAX_FRAMESIZE);
	audiodriver *self = (audiodriver *)data;
	float* buffer = (float*)output;
	float* op = buffer;
    float* ip[] = { self->workbufferL, self->workbufferR };
    if (self->isRecording()) {
        i2s(ip, op, frameCount);
    }

	self->worker->workStereo(ip, frameCount);
	//float *ip[] = { workbuffer[0], workbuffer[1] };
	float f;
	for (int i=0; i<frameCount; i++) {
		f=*ip[0]++;
		if (f>1) f=1.0f;
		if (f<-1) f=-1.0f;
		*op++ = f;
		f=*ip[1]++;
		if (f>1) f=1.0f;
		if (f<-1) f=-1.0f;
		*op++ = f;
	}
	return 0;
}

audiodriver::~audiodriver()
{
	destroyDevice();
	if (audio) {
		delete audio;
		audio = 0;
	}
}


int audiodriver::getApiDevices(int apiId) {
	try {
        audio = new RtAudio((RtAudio::RtAudioApi)apiId);
	} catch (RtError &error) {
		error.printMessage();
		return -1;
	}	

    // rtaudio returns an invalid deviceInfo for non-probed devices after a stream is opened, so probe all devices now
    for (int i = 0; i<audio->getDeviceCount(); i++) {
        audiodevice ad;
	    RtAudioDeviceInfo info;
	    try {
		    info = audio->getDeviceInfo(i+1);

            std::string deviceName = info.name;

            // ds returns devices that has duplex support twice so we remove the last one
            int j = getDeviceByName(deviceName.c_str());
            if (j != -1) continue;

            if (info.isDefault && defaultDevice == -1)
                defaultDevice = devices.size();

            ad.name = deviceName;
            ad.api_id = apiId;
            ad.device_id = i+1;
            ad.out_channels = info.outputChannels;
            ad.in_channels = info.inputChannels;
            ad.rates = info.sampleRates;
            devices.push_back(ad);

	    } catch (RtError &error) {
		    error.printMessage();
		    continue;
	    }
    }
    delete audio;
    audio = 0;
    return 0;
}


void audiodriver::initialize(audioworker *worker)
{
	this->worker = worker;
#if defined(__WINDOWS_ASIO__)
    getApiDevices(RtAudio::WINDOWS_ASIO);
#endif
#if defined(__WINDOWS_DS__)
    getApiDevices(RtAudio::WINDOWS_DS);
#endif
#if defined(__LINUX_ALSA__)
    getApiDevices(RtAudio::LINUX_ALSA);
#endif
#if defined(__LINUX_OSS__)
    getApiDevices(RtAudio::LINUX_OSS);
#endif
#if defined(__LINUX_JACK__)
    getApiDevices(RtAudio::LINUX_JACK);
#endif
#if defined(__MACOSX_CORE__)
      getApiDevices(RtAudio::RtAudio::MACOSX_CORE);
#endif
}

void audiodriver::reset()
{
	enable(false);
	enable(true);
}

bool audiodriver::enable(bool e)
{
	if (!audio)
		return false;
	if (e)
	{
        if (!started)
    		audio->startStream();
		started = true;
		return true;
	}
	else
	{
        if (started)
		    audio->stopStream();
		started = false;
		return true;
	}
}

int audiodriver::getWritePos()
{
	return 0;
}

int audiodriver::getPlayPos()
{
	return 0; 
}

void audiodriver::configure()
{
}

int audiodriver::getDeviceCount()
{
	return devices.size();
}

std::string audiodriver::getDeviceName(int index)
{
	if (index == -1)
	  index = defaultDevice; //return "invalid";
    if (index >= devices.size()) return "invalid0";
    return devices[index].name;
}


bool audiodriver::isSampleRateSupported(int index, unsigned int rate) {
    if (index == -1) return false;
    if (index >= devices.size()) return false;

    audiodevice& device = devices[index];
	
	std::vector<int>::iterator sr;
	for (sr = device.rates.begin(); sr != device.rates.end(); ++sr) {
		if (*sr == rate)
			return true;
	}

	return false;
}

bool audiodriver::isOutputDevice(int index) {
	if (index == -1) return false;
    if (index >= devices.size()) return false;
    return (devices[index].out_channels > 0)?true:false;
}

bool audiodriver::isInputDevice(int index) {
	if (index == -1) return false;
    if (index >= devices.size()) return false;
    return (devices[index].in_channels > 0)?true:false;
}

int audiodriver::getDeviceByName(const char* name) {
    for (int i=0; i<devices.size(); i++) {
        if (devices[i].name == name) return i;
    }
    return -1;
}


bool audiodriver::createDevice(int index, int inIndex)
{
	if (index == -1)
		index = getBestDevice();

	if (index == -1) return false;
	if (index >= devices.size()) return false;

	assert(desired_framecount <= MAX_FRAMESIZE);
	
	printf("creating output device '%s' with %iHz samplerate\n", getDeviceName(index).c_str(), desired_framerate);

    audio = new RtAudio((RtAudio::RtAudioApi)devices[index].api_id);

    // ensure rtaudio out/in ids are on the same api or disable input
    int outdevid = devices[index].device_id;
    int outapi = devices[index].api_id;

    int inapi = -1;
    if (inIndex != -1)
        inapi = devices[inIndex].api_id;
    int indevid = 0;
    int indevch = 0;
    if (inapi == outapi && inIndex != -1) {
        indevid = devices[inIndex].device_id;
        indevch = 2;
    }
	try {
		audio->openStream(outdevid, 2, indevid, indevch, RTAUDIO_FLOAT32, desired_framerate, (int*)&desired_framecount, 4);
		audio->setStreamCallback(&rtaudio_process_callback, (void *)this);
        if (indevid != 0)
            recording = true;
		return true;
	} catch (RtError &error) {
		error.printMessage();
	}

	return false;
}

void audiodriver::destroyDevice()
{
	if (!audio)
		return;

	enable(false);
    delete audio;
    audio = 0;
    recording = false;
}

int audiodriver::getBestDevice()
{
    return defaultDevice;
}

double audiodriver::getCpuLoad() {
    return 1.0;
}

bool audiodriver::isRecording() {
    return recording;
}

} // namespace zzub
