/*
Copyright (C) 2003-2007 Anders Ervik <calvin@countzero.no>

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "common.h"
#if defined(_MSC_VER)
// structured exception handling in visual c++
#include <eh.h>
#endif
#include <float.h>
#include <cmath>
#include "pattern.h"
#include "bmxreader.h"
#include "bmxwriter.h"
#include "tools.h"
#include "recorder.h"
#include "archive.h"

using namespace std;

/*
// unused
void copyAndScanMax(float** machineBuffer, float** pout, float numSamples, int channels, float& maxL, float& maxR) {
	maxL=0;
	maxR=0;
	for (int i=0; i<numSamples; i++) {
		float L=pout[0][i];
		if (L>maxL)
			maxL=L;
		machineBuffer[0][i]=L;
		if (channels==2) {
			float R=pout[1][i];
			machineBuffer[1][i]=R;
			if (R>maxR)
				maxR=R;
		} else
			maxR=maxL;
	}
}
*/


// this works directly like a vu-meter, declining with power over time
// maxL and maxR should be previous values, so declining of sound power is correct
inline void scanPeakStereo(float** machineBuffer, int numSamples, float& maxL, float& maxR, float falloff) {
	float *l = machineBuffer[0];
	float *r = machineBuffer[1];
	while (numSamples--) {
		maxL *= falloff;
		maxR *= falloff;
		maxL = std::max(std::abs(*l), maxL);
		maxR = std::max(std::abs(*r), maxR);
		l++;r++;
	}
}

inline void scanRMSStereo(float** machineBuffer, int numSamples, float& maxL, float& maxR) {
	maxL=0;
	maxR=0;
	float count = float(numSamples);
	float *l = machineBuffer[0];
	float *r = machineBuffer[1];
	while (numSamples--) {
		maxL += (*l)*(*l);
		maxR += (*r)*(*r);
		l++;r++;
	}
	maxL = sqrt(maxL/count);
	maxR = sqrt(maxR/count);
}

size_t findNoteParameter(zzub::patterntrack* track) {
    size_t col=zzub::no_column;
	for (size_t j=0; j<track->getParams(); j++) {
		if (track->getParam(j)->type==zzub::parameter_type_note) {
			col=j;
			break;
		}
	}

	return col;
}

namespace zzub {

/*! \class metaplugin
    \brief Private machine implementation.

    Machines see this class as metaplugin, and implements the internal
	methods of a machine. This class is only used internally. See metaplugin
	for the public interface intended for clients.
*/

/*! \class metaplugin
    \brief Base public machine implementation.

    Invoking methods on this class is thread safe.
*/


/*! \class CMachineImpl
    \brief Public machine implementation.
    \param CB Template parameter CB is a callback class.

    Invoking methods on this class is thread safe.
*/


ParameterState::ParameterState() {
	state=0;
	stateOrig=0;
	resetValues=0;
	stateLast=0;
}

ParameterState::~ParameterState() {
	if (state)
		delete state;
	if (stateOrig)
		delete stateOrig;
	if (resetValues)
		delete resetValues;
	if (stateLast)
		delete stateLast;
    
	resetValues=0;
}

void ParameterState::initialize(void* data, size_t group, size_t track, const std::vector<const zzub::parameter *>& schema) {
	if (state!=0) return ;	// only initialize once
	state=new patterntrack(group, track, schema, 1);
	stateOrig=new patterntrack((char*)data, group, track, schema, 1);
	stateLast=new patterntrack(group, track, schema, 1);

	resetValues=new patterntrack(group, track, schema, 1);
	for (size_t i=0; i<schema.size(); i++) {
		const parameter* param=resetValues->getParam(i);
		resetValues->setValue(0, i, getNoValue(param));
	}
}

// clearparameters is called so often, it was optimized like this to reduce cpu noticeably on bigger songs
void ParameterState::clearParameters() {
	if (state==0) return;

	char* srcBuf=resetValues->getTrackBuffer();
	char* destBuf=stateOrig->getTrackBuffer();
	size_t colSize=resetValues->getSize();
	memcpy(destBuf, srcBuf, colSize);
	//stateOrig->resetValues();	// the slow way
}

void ParameterState::stopParameters() {
	if (state==0) return;
	stateOrig->stopValues();
}

void ParameterState::defaultParameters() {
	if (state==0) return;
	stateOrig->defaultValues();
}

// machine parameters describing connections - used with connection tracking

std::vector<const parameter *> connectionParameters;

struct _connVolume : parameter {
	_connVolume() {
		set_word()
		.set_name("Volume")
		.set_description("Volume (0=0%, 4000=100%)")
		.set_value_min(0).set_value_max(0x4000)
		.set_value_none(0xffff)
		.set_state_flag()
		.set_value_default(0x4000)
		.append(connectionParameters);
	}
} connVolume;

struct _connPanning : parameter {
	_connPanning() {
		set_word()
		.set_name("Panning")
		.set_description("Panning (0=left, 4000=center, 8000=right)")
		.set_value_min(0).set_value_max(0x8000)
		.set_value_none(0xffff)
		.set_state_flag()
		.set_value_default(0x4000)
		.append(connectionParameters);
	}
} connPanning;

connection::connection() {
	amp=pan=0x4000;
	plugin_in=plugin_out=0;
}

metaplugin::metaplugin(zzub::player* pl, pluginloader* loader) {
	for (int c = 0; c < 2; ++c) {
		mixBuffer[c] = new float[MAXBUFFERSAMPLES*4];
		memset(mixBuffer[c], 0, sizeof(float) * MAXBUFFERSAMPLES*4);
		machineBuffer[c] = new float[MAXBUFFERSAMPLES*4];
		memset(machineBuffer[c], 0, sizeof(float) * MAXBUFFERSAMPLES*4);
	}
	sampleswritten = 0;
	machineCallbacks=new host(this);
	_internal_globalState=0;
	_internal_trackState=0;
	_name=0;

    this->player=pl;
	this->loader=loader;

	machine=0;
	tracks=0;
	x=0;
	y=0;
	lastPositionInSequencerPatternList=-1;
	lastScrollInParameterView=0;

	hardMuted=false;
	softMuted=false;
	bypassed=false;
    initialized=false;

	workTime=0;
	lastWorkPos=-1;
	lastWorkState=false;
	
	lastFrameMaxL = 0.0;
	lastFrameMaxR = 0.0;

	minimized=false;
	
//	waveFile = 0;
	writeWave = false;
	startWritePosition = -1;
	endWritePosition = -1;
	autoWrite = false;
	ticksWritten = 0;
    pluginRecorder = 0;
}

metaplugin::~metaplugin() { // todo: release pattern memory and connections
	
	clear();

	// if inConnections.size()>0 har vi ikke releaset alle connections riktig
	assert(inConnections.size()==0);
    assert(connectionStates.size()==0);

    postProcessors.clear();

    for (size_t i=0; i<tracks; i++) {
        delete trackStates[i];
    }
    trackStates.clear();

	events.clear();

	//if (implementation) delete implementation;	// denne blir deletet i mdk-bitene som er kompilert inn i maskinene

	if (machine) machine->destroy();
		
	for (int c = 0; c < 2; ++c) {
		delete[] mixBuffer[c];
		delete[] machineBuffer[c];
	}
} 


bool metaplugin::create(char* inputData, int dataSize) {

    // copy info and create plugin
    machineInfo=loader->getInfo();
	assert(machineInfo);
    machine=loader->createMachine();
    if (!machine) return false;

    // initialize plugin
    machine->_master_info=player->getMasterInfo();
    machine->_host=machineCallbacks;

    defaultAttributes();    // defaultAttributes() before init() makes some stereo wrapped machines work - instead of crashing in first attributesChanged

    if (inputData) {
		mem_archive arc;
		arc.get_outstream("")->write(inputData, dataSize);
        machine->init(&arc);
    } else {
        machine->init(0);
    }

    // set up states
	globalState.initialize(machine->global_values, 1, 0, machineInfo->global_parameters);
	_internal_globalState=globalState.getStateTrackCopy()->getValuePtr(0,0);

	setTracks(machineInfo->min_tracks);
	defaultParameters(); 

    // tell everyone we created a new plugin
    if (player->master) {
        zzub_event_data eventData={event_type_new_plugin};
        eventData.new_plugin.plugin=(zzub_plugin_t*)this;
        player->master->invokeEvent(eventData, true);
    }

	return true;
}

void metaplugin::initialize(int* attributeValues, int attributeCount, int* globalValues, int* trackValues, int tracks) {

    // initialize attributes
    defaultAttributes();
    if (attributeValues && hasAttributes()) {
        for (int i=0; i<attributeCount; i++) {
            if (i<getAttributes())
                setAttributeValue(i, attributeValues[i]);
        }
        attributesChanged();
    }

    // initialize parameters
    if (globalValues || trackValues)
        defaultParameters();

    // initialize global parameters
    if (globalValues) {
        for (size_t i=0; i<loader->plugin_info->global_parameters.size(); i++) {
            setParameter(1, 0, i, globalValues[i], false);
        }
    }

    // initialize track parameters
    if (tracks && trackValues) {
        setTracks(tracks);

        size_t trackParamCount=loader->plugin_info->track_parameters.size();
        for (int i=0; i<tracks; i++) {
            for (size_t j=0; j<trackParamCount; j++) {
                setParameter(2, i, j, trackValues[i*trackParamCount+j], false);
            }
        }
    }

    // silence any notes being initialized
    stopParameters();

    // tick and set initialized to true so processing will begin
    machine->process_events();
    initialized=true;
    
}

const std::string& metaplugin::getName() {
	return machineName;
}

void metaplugin::setName(std::string name) { 
    machineName=name; 
}

std::string metaplugin::getLoaderName() {

    return loader->getName(); 
}

void metaplugin::setTracks(size_t newTracks) {
	player->lock();
	if (newTracks>getMachineMaxTracks())
		newTracks=getMachineMaxTracks();
	if (newTracks<getMachineMinTracks())
		newTracks=getMachineMinTracks();

//	if (newTracks==getTracks()) return ;	// this test crashes Jeskola Noise XP in first tick

	if (machine)
		machine->set_track_count((int)newTracks);

	// ensure internal connection and pattern states updates with number of tracks

    size_t prevTracks=trackStates.size();
	tracks=newTracks;
	for (size_t i=0; i<patterns.size(); i++) {
		pattern* ptn=patterns[i];
		ptn->setTracks(tracks);
	}

    // if we remove tracks, free memory for the states
    for (size_t i=tracks; i<prevTracks; i++)
        delete trackStates[i];

    trackStates.resize(tracks);

    size_t trackSize = machineInfo->get_group_size(2);

	for (size_t i=prevTracks; i<tracks; i++) {
		trackStates[i]=new ParameterState();
		trackStates[i]->initialize((char*)machine->track_values + i*trackSize, 2, (int)i, machineInfo->track_parameters);

        // copy novalue states internally:
		trackStates[i]->copyChangedParameters();

        // set defaults so player can set after unlocking:
        trackStates[i]->defaultParameters();
        trackStates[i]->stopParameters();   // bass 3 has multiple note columns and wont play unless note columns are initialized to a stop
	}

	// populate real metaplugin-members as expected read by machines using hack.h
	if (tracks>0)
		_internal_trackState=trackStates[0]->getStateTrackCopy()->getValuePtr(0,0); else
        _internal_trackState=0;

	player->unlock();
}


sequence* metaplugin::createSequence() {
	return player->song_sequencer.createTrack(this);
}

void metaplugin::addPattern(pattern* pattern) {
	patterns.push_back(pattern);
}

bool metaplugin::removePattern(size_t index) {
	player->lockTick();
	pattern* p=patterns[index];
	patterns.erase(patterns.begin()+index);

	// Remove sequence references as well

	for (size_t i=0; i<getPlayer()->getSequenceTracks(); i++) {
		sequence* t=getPlayer()->getSequenceTrack(i);
		if (t->getMachine()==this) {
			t->removeEvents(p);
			if (t->pattern == p)
				t->pattern = 0;
		}
	}

	delete p;
	player->unlockTick();
	return true;
}

bool metaplugin::movePattern(size_t index, size_t newIndex) {
    if (patterns.size()<2) return false;
	if (newIndex>patterns.size()-1) return false;

    vector<pattern*> newPatterns(patterns.size());
	int pos=0;
	for (size_t i=0; i<patterns.size(); i++) {
		if (i==newIndex) {
			newPatterns[i]=patterns[index];
		}  else {
			if (i==index && newIndex>index) 
				pos++;
			newPatterns[i]=patterns[pos];
			if (i==index && newIndex<index) 
				pos++;
			pos++;
		}
	}
	patterns=newPatterns;
	return true;

}


pattern* metaplugin::createPattern(size_t rows) {
	pattern* ptn = new pattern(machineInfo, getConnections(), getTracks(), rows);
	patterns.push_back(ptn);
	return ptn;
}

// clears everything but events - not sure if thats good, but we dont have to reassign master events. events should be cleaned by the client anyway
void metaplugin::clear() {

	while (getConnections()) {
		connection* conn=getConnection((size_t)0);
		if (!conn) break;
		deleteInput(conn->plugin_in);
	}

	defaultParameters();

	// test if any machines are connected to this machine
	for (size_t i=0; i<player->getMachines(); i++) {
		metaplugin* m=player->getMachine(i);
		for (size_t j=0; j<m->getConnections(); j++) {
			connection* conn=m->getConnection(j);
			if (!conn) continue;
			if (conn->plugin_in==this)			// dette er vel ikke helt som planlagt? burde ikke j bli decreaset?
				conn->plugin_out->deleteInput(conn->plugin_in);
		}
	}

	for (size_t i=0; i<patterns.size(); i++) {
		delete patterns[i];
	}
	patterns.clear();
	
	// check if waveFile still open, if yes close it
    if (pluginRecorder && pluginRecorder->isOpen())
        pluginRecorder->close();

}

void metaplugin::clearUnChangedParameters() {
	globalState.clearUnChangedParameters();
	for (size_t i=0; i<trackStates.size(); i++) {
		trackStates[i]->clearUnChangedParameters();
	}

}

void ParameterState::clearUnChangedParameters() {
	for (size_t i=0; i<stateOrig->getParams(); i++) {
		int v=stateOrig->getValue(0, i);
		int c=stateLast->getValue(0, i);

		int nv=getNoValue(stateOrig->getParam(i));
		DWORD flags=stateOrig->getParam(i)->flags;
		if (v!=nv && v==c) {
			stateOrig->setValue(0, i, nv);
		}
	}
}

void metaplugin::copyChangedParameters() {
    // 19. oct 2006, lets try to do this do connections as well and see what happens - i hope this allows saving correct pan/amp defaults in zxms- oh yes!
    for (size_t i=0; i<getConnections(); i++)
        connectionStates[i]->copyChangedParameters();

	globalState.copyChangedParameters();
	for (size_t i=0; i<getTracks(); i++)
		trackStates[i]->copyChangedParameters();
}

void ParameterState::copyChangedParameters() {
	for (size_t i=0; i<stateOrig->getParams(); i++) {
		int c=state->getValue(0, i);
		int v=stateOrig->getValue(0, i);

		int nv=getNoValue(state->getParam(i));
		if (v!=nv)
			state->setValue(0, i, v);
		
		stateLast->setValue(0, i, v);
	}
}

void metaplugin::tickAsync() {
    if (!initialized) return ;

	clearUnChangedParameters();
	
	_SEH_TRY
	machine->process_events();
	
	copyChangedParameters();

	_SEH_CATCH("Exception %x in tick: %s", getLoaderName().c_str());

}

void metaplugin::postProcessEvents() {
    for (size_t i=0; i<postProcessors.size(); i++) {
        tickstream* ts = postProcessors[i];
        ts->process_events();
    }
}

void metaplugin::postProcessStereo(bool zero, int numSamples) {
    for (size_t i=0; i<postProcessors.size(); i++) {
        tickstream* ts = postProcessors[i];
        ts->process_stereo(machineBuffer, numSamples);
    }
}

void metaplugin::addPostProcessor(tickstream* ts) {
    postProcessors.push_back(ts);
}

bool metaplugin::removePostProcessor(tickstream* ts) {
    vector<tickstream*>::iterator i =find(postProcessors.begin(), postProcessors.end(), ts);
    if (i == postProcessors.end()) return false;
    postProcessors.erase(i);
    return true;
}



void metaplugin::tick() {
	//if (!machine) return ;    // doesnt happen
    if (!initialized) return ;

	clearUnChangedParameters();

	player_state state=player->getPlayState();
	if (state==player_state_playing)
		player->currentlyPlayingSequencer->advanceMachine((metaplugin*)this);

	_SEH_TRY
	machine->process_events();
	lastTickPos=player->getWorkPosition();
	
    postProcessEvents();
	if (autoWrite) {
		if (state==player_state_playing)
			writeWave = true;
		else
			writeWave = false;
	}
	
	if (writeWave && (endWritePosition != -1) && (lastTickPos >= endWritePosition)) {
		writeWave = false;
	}
	if (!writeWave && (startWritePosition != -1) && (lastTickPos >= startWritePosition)) {
		writeWave = true;
	}
	
	if (writeWave) {
		ticksWritten++;
	}
	
	copyChangedParameters();

	_SEH_CATCH("Exception %x in tick: %s", getLoaderName().c_str());

}

zzub::player* metaplugin::getPlayer() {
	return (zzub::player*)player;
}

std::string metaplugin::describeValue(size_t param, int value) {
	const char* desc=machine->describe_value(param, value);
	if (!desc) return "";
	return desc;
}

std::string metaplugin::describeValue(size_t group, size_t param, int value) {
	if (group == 0) {
		char pc[16];
		// buzz writes this as "-X.Y dB (Z%)"
		sprintf(pc, "%04x", value);
		return pc;
	}
	size_t currentParameter=no_column;
    pattern conv(machineInfo, getConnections(), getTracks(), 0);
	if (conv.patternToLinear(group, 0, param, currentParameter))
	    currentParameter-=getConnections()*2;	// subtract connection parameters

	const parameter* para=getMachineParameter(group, 0, param);
	if (currentParameter!=no_column) {
		if (value!=getNoValue(para)) {	// infector crashen when trying to describe novalues (and out-of-range-values)
			return describeValue(currentParameter, value);
		}
	}
	return "";
}


void metaplugin::command(int cmd) {
	// lr: removed locks since some commands open
	// modal dialogs and do not return, e.g. PVST.
	// the plugin has to lock the player if it needs to.
	//player->lock();
	machine->command(cmd);
	//player->unlock();
}

void metaplugin::attributesChanged() {
	machine->attributes_changed();
}

size_t metaplugin::getAttributes() {
	if (machineInfo->attributes.size()<0) return 0;
	return (size_t)machineInfo->attributes.size();
}

int metaplugin::getAttributeValue(size_t i) {
	return machine->attributes[i];
}

const attribute& metaplugin::getAttribute(size_t i) {
	return *machineInfo->attributes[i];
}

// used by buzzreader to determine if pointer is writable - some machines may say they have attributes but not provide a pointer (?)
bool metaplugin::hasAttributes() {
	return machine->attributes!=0;
}

void metaplugin::setAttributeValue(size_t i, int value) {
	machine->attributes[i]=value;
}

bool metaplugin::doesInputMixing() {
	return ((machineInfo->flags&plugin_flag_does_input_mixing)!=0);
}

float** metaplugin::getBuffer(DWORD* lastBufferSize) {
	if (lastBufferSize)
		*lastBufferSize=lastWorkSamples;
	return machineBuffer;
}


void metaplugin::save(StreamWriter& writer) {
	mem_archive arc;
	machine->save(&arc);
	CBuzzOutput output(writer);
	std::vector<char> &b = arc.get_buffer("");
	if (b.size()) {
		output.write(&b[0], b.size());
	}
}


void metaplugin::stop() {
	machine->stop();
}

void metaplugin::softMute() {
	clearParameters();
	softMuted=true;
}

void metaplugin::mute(bool state) {
	hardMuted=state;
}

void metaplugin::bypass(bool state) {
	bypassed=state;
}


pattern* metaplugin::getPattern(std::string name) {
	for (size_t i=0; i<patterns.size(); i++) {
		if (patterns[i]->name==name) return patterns[i];
	}
	return 0;
}

int metaplugin::getPatternIndex(pattern* p) {
	for (size_t i=0; i<patterns.size(); i++) {
		if (patterns[i]==p) return (int)i;
	}
	return -1;
}

std::string metaplugin::getNewPatternName() {
	char pc[16];

	for (int i=0; i<9999; i++) {
		if (i<100)
			sprintf(pc, "%.02i", i); else
		if (i<1000)
			sprintf(pc, "%.03i", i); else
		if (i<10000)
			sprintf(pc, "%.04i", i); else
		if (i<100000)
			sprintf(pc, "%.05i", i);
		pattern* p=getPattern(pc);
		if (!p) return pc;
	}
	return "NaN";
}

bool isCircular(metaplugin* inputMachine, metaplugin* outputMachine) {
	if (inputMachine==outputMachine) return true;

	for (size_t i=0; i<inputMachine->getConnections(); i++) {
		metaplugin* inputInput=inputMachine->getConnection(i)->plugin_in;
		bool ic=isCircular(inputInput, outputMachine);
		if (ic) return true;
	}
	return false;
}

bool metaplugin::addInput(metaplugin* fromMachine, unsigned short amp, unsigned short pan) {
	// Ensure this connection won't crash
	// maskiner med MIF_NO_OUTPUT kan kun kobles mot master, faktisk M de alltid kobles mot master automatisk

	bool isNoOutput=false;

	if ((fromMachine->machineInfo->flags&plugin_flag_no_output) ) isNoOutput=true;
	if (isNoOutput && getType()!=plugin_type_master) return false;

	// Check whether these machines are already connection to each others
	if (getConnection(fromMachine)) return false;

    // check cyclic connections
	if (isCircular(fromMachine, this)) return false;

	// generators don't have inputs, but some popular plugins are incorrectly flagged as generators
    // the following line was commented out in response to that, so f.ekx songs using geoniks 2p filter will load
	if (getType()==plugin_type_generator) return false;
    if (fromMachine->getType()==plugin_type_master) return false;

	// Create the connection:

	connection* conn=new connection();
	conn->amp=amp;
	conn->pan=pan;
	conn->plugin_in=fromMachine;
	conn->plugin_out=this;

	player->lock();

	// the following adds NO_OUTPUT-machines always to the beginning of master connection chain
	// so f.ex aux machines are always worked first
	if (!isNoOutput)
		inConnections.push_back(conn); else
		inConnections.insert(inConnections.begin(), conn);
	fromMachine->outConnections.push_back(conn);

	// unknown whether we should call CMachineInterfaceEx::AddInput for inputmixers only
	//if (doesInputMixing())
	machine->add_input(fromMachine->getName().c_str());

	for (size_t i=0; i<patterns.size(); i++) {
		patterns[i]->addInput();
	}

    // create connection states
	ParameterState* state=new ParameterState();
	state->initialize(conn, 0, inConnections.size()-1, connectionParameters);
	connectionStates.push_back(state);

	player->unlock();

    zzub_event_data eventData={event_type_connect};
    eventData.connect_plugin.connection=(zzub_connection_t*)conn;
	invokeEvent(eventData, true);

	return true;
}

void metaplugin::deleteInput(metaplugin* fromMachine) {
	using namespace std;

	player->lock();

    // notify plugin about deletion
	machine->delete_input(fromMachine->getName().c_str());

	connection* conn=0;	// store a pointer we can delete

    for (size_t i=0; i<inConnections.size(); i++) {
        if (conn) {
            connectionStates[i]->getStateTrack()->setTrack(i);
            connectionStates[i]->getStateTrackCopy()->setTrack(i);
        }
        if (!conn && inConnections[i]->plugin_in==fromMachine) {
        	// delete connection patterns
	        for (size_t j=0; j<patterns.size(); j++) {
                patterns[j]->deleteInput(i);
	        }

            // delete state
            ParameterState* connectionState=connectionStates[i];
            delete connectionState;

            conn=inConnections[i];

            // remove entry from collections
            connectionStates.erase(connectionStates.begin()+i);
            inConnections.erase(inConnections.begin()+i);

        }
    }

    // remove from plugins output collection
    for (size_t i=0; i<fromMachine->outConnections.size(); i++) {
        if (fromMachine->outConnections[i]->plugin_out==this) {
            fromMachine->outConnections.erase(fromMachine->outConnections.begin()+i);
            break;
        }
    }

	player->unlock();

    zzub_event_data eventData={event_type_disconnect};
    eventData.disconnect_plugin.connection=(zzub_connection_t*)conn;
	invokeEvent(eventData, true);

	delete conn;

}


void metaplugin::clearParameters() {
	// dont reset connection parameters

	globalState.clearParameters();

	for (size_t i=0; i<tracks; i++) {
		trackStates[i]->clearParameters();
	}
}

// stopParameters kalles kun i initialisering - kanskje vi kan kjre den p en vanlig stop() ogs?
void metaplugin::stopParameters() {

	globalState.stopParameters();

	for (size_t i=0; i<getTracks(); i++) {
		trackStates[i]->stopParameters();
	}
}


void metaplugin::defaultAttributes() {
	int attributeCount=machineInfo->attributes.size();

	for (size_t k=0; k<this->getAttributes(); k++) {
		if (machine->attributes) {
			machine->attributes[k]=machineInfo->attributes[k]->value_default;
		}
	}
} 

void metaplugin::defaultParameters() {
	if (machine==0) return ;

	if (machine->global_values!=0)
		globalState.defaultParameters();

	for (size_t i=0; i<getTracks(); i++) {
		if (i<trackStates.size())
			trackStates[i]->defaultParameters();
	}

}

patterntrack* metaplugin::getStateTrack(size_t group, size_t index) {
	switch (group) {
		case 0:
			if (index>=connectionStates.size()) return 0;
			return connectionStates[index]->getStateTrack();
		case 1:
			return globalState.getStateTrack();
		case 2:
			if (index>=trackStates.size()) return 0;
			return trackStates[index]->getStateTrack();
		default:
			return 0;
	}
}

patterntrack* metaplugin::getStateTrackCopy(size_t group, size_t index) {
	switch (group) {
		case 0:
            if (index>=connectionStates.size()) return 0;
			return connectionStates[index]->getStateTrackCopy();
		case 1:
			return globalState.getStateTrackCopy();
		case 2:
			if (index>=trackStates.size()) return 0;
			return trackStates[index]->getStateTrackCopy();
		default:
			return 0;
	}
}

// getCommands returns a blank string when there are no commands, otherwise a \n-separated string
std::string metaplugin::getCommands() {
	if (!machineInfo->commands) return "";
	return machineInfo->commands;
}

// if a command string starts with the char '\', it has subcommands
// unexpectedly, this returns a \n-separated string (like getCommands())
// some machines need to be ticked before calling getSubCommands (not yet supported)
std::string metaplugin::getSubCommands(int i) {
	StringWriter wr;
	CBuzzOutput output(wr);

	machine->get_sub_menu(i, &output);
	if (wr.position()==0) return "";
	wr.write((char)0);	// terminate array

	// create a new \n-separated string and return it instead, means both getCommands() and getSubCommands() return similar formatted strings
	const char* firstp=wr.getString();
    string ret="";

	while (*firstp) {
		if (ret.length()>0)
            ret+="\n";
        ret+=firstp;
		firstp+=strlen(firstp)+1;
	}
	return ret;
}


// this takes values from of a specified row on a pattern and plays it
void metaplugin::playPatternRow(pattern* pattern, size_t row, bool record) {

	// TODO: use setParameter for vol/pan as well (n er ParameterStates implementert for vol/pan, s dette skal funke - m verifiseres)
	for (size_t i=0; i<getConnections(); i++) {
		connection* conn=getConnection(i);
		patterntrack* connTrack=pattern->getPatternTrack(0, i);
        for (size_t j=0; j<connTrack->getParams(); j++) {
            int value=connTrack->getValue(row, j);
			if (value!=getNoValue(connTrack->getParam(j)))
				setParameter(0, i, j, value, record);

        }
	}

	patterntrack* globTrack=pattern->getPatternTrack(1, 0);
	for (size_t j=0; j<globTrack->getParams(); j++) {
		int value=globTrack->getValue(row, j);
		if (value!=getNoValue(globTrack->getParam(j)))
			setParameter(1, 0, (int)j, value, record);
	}

	if (machine->track_values) {
		size_t trackSize = machineInfo->get_group_size(2);
		for (size_t i=0; i<getTracks(); i++) {
			patterntrack* track=pattern->getPatternTrack(2, i);
			if (track==0) continue;	// thread safety =)
			
			for (size_t j=0; j<track->getParams(); j++) {
				int value=track->getValue(row, j);
				if (value!=getNoValue(track->getParam(j)))
					setParameter(2, i, j, value, record);
			}

		}
	}
	softMuted=false;
}

void metaplugin::recordParameter(size_t group, size_t track, size_t param, int value) {
	sequencer* seqr=player->currentlyPlayingSequencer;
	for (size_t i=0; i<seqr->getTracks(); i++) {
		sequence* seq=seqr->getTrack(i);
		pattern* p=seq->getCurrentlyPlayingPattern();
		if (!p) continue;

		patterntrack* t=p->getPatternTrack(group, track);
		if (!t) continue;

		size_t row=seq->getCurrentlyPlayingPatternPosition();
		if (row>=p->getRows()) row=p->getRows()-1;
		t->setValue(row, param, value);
	}
}


void metaplugin::setParameter(size_t group, size_t track, size_t param, int value, bool record) {

	patterntrack* pt=getStateTrack(group, track);
	if (pt==0) return ;

	pt->setValue(0, param, value);

	if (record && player->recordParameters)
		recordParameter(group, track, param, value);

	// send notification
	zzub_event_data eventData={ event_type_parameter_changed };
	eventData.change_parameter.group=group;
	eventData.change_parameter.track=track;
	eventData.change_parameter.param=param;
	eventData.change_parameter.value=value;
	invokeEvent(eventData);

	softMuted=false;
}

int metaplugin::getParameter(size_t group, size_t track, size_t param) {

	// get most recent value first
	patterntrack* pt=getStateTrack(group, track);
	if (pt==0) return 0;	// TODO: should use use default value

	int value=pt->getValue(0, param);

	// if recent value is NoValue, return last known value
	if (value==getNoValue(pt->getParam(param))) {
		patterntrack* pc=getStateTrackCopy(group, track);
		if (!pc) return value;
		return pc->getValue(0, param);
	}

	return value;
}

const parameter* metaplugin::getMachineParameter(size_t group, size_t track, size_t param) {
	size_t index=0;
	pattern conv(machineInfo, getConnections(), getTracks(), 0);
	conv.patternToLinear(group, track, param, index);
	return conv.getColumnParameter(index);
}

int metaplugin::getMachineParameterValue(size_t index) {
	// find value this parameter index in machine->gvals or
	size_t linearIndex=index+getConnections()*2;
	size_t group, track, column;

	pattern conv(machineInfo, getConnections(), getTracks(), 0);
	if (!conv.linearToPattern(linearIndex, group, track, column))
		return 0;

	// NOTE: we may want to use getStateTrack here?
	patterntrack* tr=this->getStateTrackCopy(group, track);
	if (!tr) 
		return 0;
	return tr->getValue(0, column);
}

void metaplugin::findNoteColumn(size_t& noteColumn, size_t& noteGroup, bool& multitrackKeyJazz) {
	size_t found_col=no_column;
	multitrackKeyJazz=false;
	noteColumn=no_column;
	noteGroup=no_group;

	patterntrack* glob=getStateTrack(1, 0);
	if (glob) {
		found_col=findNoteParameter(glob);
		if (found_col!=no_column) {
			noteColumn=found_col;
			noteGroup=1;
		}
	}

	if (found_col!=no_column) return;

	if (getTracks()>0) {
		patterntrack* trk=getStateTrack(2, 0);
		found_col=findNoteParameter(trk);

		if (found_col!=no_column) {
			multitrackKeyJazz=true;
			noteColumn=found_col;
			noteGroup=2;
		}

	}
}

// counts how many events of a specific type a machine supports (currently only used by the default master-validator which doesnt validate)
void metaplugin::addEventHandler(event_handler *ev) {
    for (size_t i=0; i<events.size(); i++)
        if (events[i]==ev) return ;

	events.push_back(ev);
}

bool metaplugin::invokeEvent(zzub_event_data_t& data, bool immediate) {
	bool handled=false;
	for (size_t i=0; i<events.size(); i++) {
		//if (events[i]->et==data.type) {
        event_handler* ev=events[i];
        if (!immediate) {
            event_message em = {this, ev, data };
            player->eventLock.Lock();
            player->messageQueue.push_back(em);
            player->eventLock.Unlock();
        } else {
            handled=ev->invoke(data)||handled;
        }
	}
	return handled;
}

void metaplugin::removeEventHandler(event_handler* ev) {

    // flush messages for this handler
    player->eventLock.Lock();
    for (deque<event_message>::iterator i = player->messageQueue.begin(); i!=player->messageQueue.end(); i++) {
        if (i->event == ev) {
            i = player->messageQueue.erase(i);
            --i;
        }
    }
    player->eventLock.Unlock();

    for (vector<event_handler*>::iterator i=events.begin(); i!=events.end(); ++i) {
		if (*i==ev) {
			events.erase(i);
			return ;
		}
	}
}

size_t metaplugin::getTracks() { 
	if ((size_t)machineInfo->min_tracks>tracks)
		return machineInfo->min_tracks; 
	return tracks;
}

pattern* metaplugin::getPattern(size_t index) { 
	if (patterns.size()==0) 
		return 0; 
	if (index<0 || index>=patterns.size()) 
		return 0; 
	return patterns[index]; 
}

size_t metaplugin::getPatterns() { 
	return patterns.size(); 
}

size_t metaplugin::getConnections() {
	size_t size=inConnections.size();
	return size;
}

connection* metaplugin::getConnection(size_t index) {
	connection* c=0;
	if (index<inConnections.size())
		c=inConnections[index]; 
	return c;
}

connection* metaplugin::getConnection(metaplugin* machine) {
	connection* c=0;
	for (size_t i=0; i<inConnections.size(); i++) {
		if (inConnections[i]->plugin_in==machine) {
			return inConnections[i];
		}
	}
	return 0;
}

size_t metaplugin::getOutputConnections() {
	size_t size=outConnections.size();
	return size;
}

connection* metaplugin::getOutputConnection(size_t index) {
	connection* c=0;
	if (index<outConnections.size())
		c=outConnections[index]; 
	return c;
}

bool metaplugin::isSoloMutePlaying() { 
	if (isSoftMuted()) return false;
	if (isMuted()) return false;
	if (!player->getSoloMachine()) return true;
	if (getType()!=plugin_type_generator) return true;
	if (player->getSoloMachine()==this) 
		return true; else
		return false;
}

size_t metaplugin::getEnvelopeInfos() {
	envelope_info const ** ei=machine->get_envelope_infos();
	if (ei==0) return 0;
	size_t count=0;
	while (*ei) { count++; ei++; }
	return count;
}

const envelope_info& metaplugin::getEnvelopeInfo(size_t index) {
	envelope_info const ** ei=machine->get_envelope_infos();
	return *ei[index];
}

size_t metaplugin::getStateParameters() {
	size_t count=getConnections();  // volume are state parameters
	count+=globalState.getStateTrack()->getStateParams();
	for (size_t i=0; i<trackStates.size(); i++)
		count+=trackStates[i]->getStateTrack()->getStateParams();

	return count;
}

// resetMixer srger for at maskinene mixer p nytt etter stop/play (ellers fr vi asserts i player::workmachine() nr posintick==lastWorkPos==0 && mixSize!=prevSize)
void metaplugin::resetMixer() {
	if (getType()==plugin_type_master)
		machine->_master_info->tick_position=0;
	lastWorkPos=-1;
}

bool metaplugin::setInstrumentName(std::string name) {
	machine->set_instrument(name.c_str());

	return true;
}

bool metaplugin::work(float** pout, int numSamples, unsigned long flags) {
    if (!initialized) return false;

	float samplerate = float(machine->_master_info->samples_per_second);
	float falloff = std::pow(10.0f, (-48.0f / (samplerate * 20.0f))); // vu meter falloff (-48dB/s)
    bool ret=flags==process_mode_write?false:true;
    double tempTime=player->timer.frame();

	if (!isBypassed()) {
		// make a copy of our pointers so the plugin can iterate them safely
		float *plin[] = {pout[0],pout[1]};
		float *plout[] = {machineBuffer[0],machineBuffer[1]};
        _SEH_TRY
            ret=machine->process_stereo(plin, plout, numSamples, flags); 
        _SEH_CATCH("Work exception %x in %s", getLoaderName().c_str());
    }

	// make a copy so we can return the same buffer later (otherwise stereo self-mixers bug with multiple outputs)
	if (ret) {
		// copyAndScanMax(machineBuffer, pout, numSamples, workChannels, lastFrameMaxL, lastFrameMaxR);
		// scanPeak(machineBuffer, numSamples, 2, lastFrameMaxL, lastFrameMaxR);
		scanPeakStereo(machineBuffer, numSamples, lastFrameMaxL, lastFrameMaxR, falloff);
		//scanRMSStereo(machineBuffer, numSamples, lastFrameMaxL, lastFrameMaxR);
	} else {
		// apply falloff
		lastFrameMaxR *= std::pow(falloff, numSamples);
		lastFrameMaxL *= std::pow(falloff, numSamples);
	}

    postProcessStereo(ret, numSamples);
	writeWaveBuffer(ret, numSamples);

    // process scheduled events
    for (list<scheduled_event>::iterator i = scheduledEvents.begin(); i != scheduledEvents.end(); ++i) {
        if (player->workPos >= i->time) {
            // on which thread are these coming in??
            machine->event(i->data);
            i = scheduledEvents.erase(i);
            --i;
        }
    }


	// keep some data for statistics
	lastWorkState=ret;
	lastWorkPos=machine->_master_info->tick_position;
	lastWorkSamples=numSamples;
	sampleswritten += numSamples;

	if (machineInfo->type == zzub_plugin_type_master) {
		// send vu information
		zzub_event_data eventData={zzub_event_type_vu};
		eventData.vu.size = numSamples;
		eventData.vu.left_amp = lastFrameMaxL;
		eventData.vu.right_amp = lastFrameMaxR;
		eventData.vu.time = float(sampleswritten) / samplerate;
		invokeEvent(eventData, false);
	}

	workTime=player->timer.frame()-tempTime;

	return ret;
}

void metaplugin::input(float** buffer, int numSamples, float amp) {
	machine->input(buffer, numSamples, amp);
}


/*

    metaplugin column space conversion helpers

*/
bool metaplugin::stateToPatternSpace(size_t index, size_t& group, size_t& track, size_t& column) {
	size_t currentParameter=0;

	for (size_t i=0; i<getConnections(); i++) {
		if (index == currentParameter) {
			group=0;
			track=i;
			column=0;
			return true;
		}
		currentParameter++;
	}

	for (size_t i=0; i<machineInfo->global_parameters.size(); i++) {
		if (machineInfo->global_parameters[i]->flags&parameter_flag_state) {
			if (index==currentParameter) {
				group=1;
				track=0;
				column=i;
				return true;
			}
			currentParameter++;
		}
	}

	for (size_t j=0; j<getTracks(); j++) {
		for (size_t i=0; i<machineInfo->track_parameters.size(); i++) {
			if (machineInfo->track_parameters[i]->flags&parameter_flag_state) {
				if (index==currentParameter) {
					group=2;
					track=j;
					column=i;
					return true;
				}
				currentParameter++;
			}
		}
	}
	return false;
}

void metaplugin::setWriteWave(bool enable) {
	writeWave = enable;
}

bool metaplugin::getWriteWave() {
	return writeWave;
}

void metaplugin::setStartWritePosition(int position) {
	startWritePosition = position;
}

void metaplugin::setEndWritePosition(int position) {
	endWritePosition = position;
}

int metaplugin::getStartWritePosition() {
	return startWritePosition;
}

int metaplugin::getEndWritePosition() {
	return endWritePosition;
}

void metaplugin::writeWaveBuffer(bool full, int numSamples) {
	if (!pluginRecorder) return ;

    if (writeWave) { // shall we write a wavefile?
		if (!pluginRecorder->isOpen()) { // did we open the handle yet?
            //pluginRecorder->samplesPerSecond=player->masterInfo.samples_per_second;
            pluginRecorder->open();
		}
		if (pluginRecorder->isOpen()) { // do we have a handle?
			if (!full) { // buffer is not mixed
				memset(machineBuffer[0], 0, sizeof(float)*numSamples);
				memset(machineBuffer[1], 0, sizeof(float)*numSamples);
			}
            pluginRecorder->write(machineBuffer, numSamples);
		}
	} else { // no wave writing
		if (pluginRecorder->isOpen()) { // but our handle is still open
            pluginRecorder->close();
		}
	}
}

void metaplugin::setAutoWrite(bool enabled) {
	autoWrite = enabled;
}

bool metaplugin::getAutoWrite() {
	return autoWrite;
}

int metaplugin::getTicksWritten() {
	return ticksWritten;
}

void metaplugin::resetTicksWritten() {
	ticksWritten = 0;
}

void metaplugin::setRecorder(zzub::recorder* r) {
    pluginRecorder = r;
}

zzub::recorder* metaplugin::getRecorder() {
    return pluginRecorder;
}


} // namespace zzub
