/*
Copyright (C) 2003-2007 Anders Ervik <calvin@countzero.no>

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "common.h"
#include <functional>
#include <algorithm>
#include <time.h>
#include <float.h>
#include "bmxreader.h"
#include "bmxwriter.h"
#include "timer.h"
#include "dummy.h"

#include "tools.h"

#if defined(POSIX)
#include <dirent.h>
#include <sys/stat.h>
#endif

#if defined(__SSE__)
// denormalisation issues can be avoided by turning on
// flush zero mode, which is an SSE instruction.
// at the moment, this is only being enabled for 
// gcc and i686 machines. modify sconstruct to 
// support other targets.
// also see http://www.intel.com/cd/ids/developer/asmo-na/eng/dc/pentium4/knowledgebase/90575.htm
	#include <xmmintrin.h>

	#if defined(__GNUC__)
		#define SETGRADUN setgradun_
		#define SETABRPUN setabrpun_
	#endif
	
	void SETGRADUN()
	{
	 _MM_SET_FLUSH_ZERO_MODE (_MM_FLUSH_ZERO_OFF);
	}

	void SETABRPUN()
	{
	 _MM_SET_FLUSH_ZERO_MODE (_MM_FLUSH_ZERO_ON);
	}
#else
#define SETGRADUN()
#define SETABRPUN()
#endif

// HKEY_CURRENT_USER\Software\Jeskola\Buzz\Settings
// her er en key som heter BuzzPath, som forteller roten av default buzz-installasjon

using namespace std;

namespace zzub {

short player::oscTables[8][OSCTABSIZE];

// generateOscillatorTables generates diverse tables used by some machines
void player::generateOscillatorTables() {
	int tabSize=2048;
	srand(static_cast<unsigned int>(time(0)));
	for (int tabLevel=0; tabLevel<11; tabLevel++) {
		int tabOfs=zzub::get_oscillator_table_offset(tabLevel);
		for (int i=0; i<tabSize; i++) {
			double dx=(double)i/tabSize;
			oscTables[zzub::oscillator_type_sine][tabOfs+i]=(short)(sin(dx*2.0f*PI)*32000);
			oscTables[zzub::oscillator_type_sawtooth][tabOfs+i]=(short)(sawtooth(dx*2.0f*PI)*32000);
			oscTables[zzub::oscillator_type_pulse][tabOfs+i]=(short)(square(dx*2.0f*PI)*32000);
			oscTables[zzub::oscillator_type_triangle][tabOfs+i]=(short)(triangle(dx*2.0f*PI)*32000);
			oscTables[zzub::oscillator_type_noise][tabOfs+i]=(short) (((float)rand()/(float)RAND_MAX)*64000.f - 32000);
			oscTables[zzub::oscillator_type_sawtooth_303][tabOfs+i]=(short)(sawtooth(dx*2.0f*PI)*32000);
			oscTables[6][tabOfs+i]=(short)(sin(dx*2.0f*PI)*32000);
		}
		tabSize/=2;
	}
}

void player::loadPluginLibrary(const std::string &fullpath) {
	int dpos=(int)fullpath.find_last_of('.');
	string fileExtension = fullpath.substr(dpos);
#if defined(_WIN32)
	transform(fileExtension.begin(), fileExtension.end(), fileExtension.begin(), toLower);
	if (fileExtension == ".dll") {	
#elif defined(POSIX)
	if (fileExtension == ".so") {
#endif
		pluginlib* lib=new pluginlib(fullpath,*this);
		// machine loaders will be registered by lib through registerMachineLoader,
		// now and during loading of songs
		pluginLibraries.push_back(lib);
	}
}

void player::registerMachineLoader(pluginloader *pl) {
	// even if the library has no plugins, we need to keep it
	// e.g. lunar will load plugins later on.
	string pluginUri=pl->getUri();
	transform(pluginUri.begin(), pluginUri.end(), pluginUri.begin(), toLower);
	if (!isBlackListed(pluginUri)) {
		machines[pluginUri]=pl;
	}
}

void player::initializeFolder(std::string folder) {
	using namespace std;

#if defined(_WIN32)

	string searchPath=folder + "*.dll";

	WIN32_FIND_DATA fd;
	HANDLE hFind=FindFirstFile(searchPath.c_str(), &fd);

	while (hFind!=INVALID_HANDLE_VALUE) {
		
		if ( (fd.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY)==0) {
			string fullFilePath=folder + fd.cFileName;
			printf("enumerating %s\n", fullFilePath.c_str());
			char fullPath[1024];
			char* filePart;
			GetFullPathName(fullFilePath.c_str(), 1024, fullPath, &filePart);

			loadPluginLibrary(fullPath);
		}

		if (!FindNextFile(hFind, &fd)) break;
	}
	FindClose(hFind);

#elif defined(POSIX)
	struct dirent **namelist;
	struct stat statinfo;
	int n;
	
	string searchPath=folder;
	
    n = scandir(searchPath.c_str(), &namelist, 0, alphasort);
    if (n < 0)
        perror("scandir");
    else {
        while(n--) {
			string fullFilePath=folder + namelist[n]->d_name;
			printf("enumerating %s\n", fullFilePath.c_str());
			if (!stat(fullFilePath.c_str(), &statinfo))
			{
				if (!S_ISDIR(statinfo.st_mode))
				{
					loadPluginLibrary(fullFilePath);
				}
			}
            free(namelist[n]);
        }
        free(namelist);
    }

#endif

}


/***

	player

***/

/*! \class player
    \brief Public interface for library clients.

    Invoking methods on this class is thread safe. Note that this class
	contains public members and methods that return objects that are not thread
	safe.
*/

/*! \brief Constructor for new player instances.

*/
player::player() {
	for (int c = 0; c < 2; ++c) {
		mixBuffer[c] = new float[MAXBUFFERSAMPLES*4];
		memset(mixBuffer[c], 0, sizeof(float) * MAXBUFFERSAMPLES*4);
	}
	soloMachine=0;
	workPos=0;
    workFracs=0.0f;
	playerState=player_state_muted;
    currentlyPlayingSequencer=&song_sequencer;
    master=0;
	masterLoader = 0;
    recordParameters=false;
    inputBuffer[0] = 0;
    inputBuffer[1] = 0;
	cpuLoad = 0;
}

player::~player(void) {
	// free machineloaders
	clearMachineLoaders();
	machineInstances.clear();	// finally remove master

	if (master)
		delete master;
	if (masterLoader)
		delete masterLoader;
	for (int c = 0; c < 2; ++c) {
		delete[] mixBuffer[c];
	}
}


void player::addMachineFolder(std::string folder) {
	pluginFolders.push_back(folder);
}

void player::clearMachineLoaders() {
    for (size_t i=0; i<pluginLibraries.size(); i++) {
        pluginlib* lib = pluginLibraries[i];
        delete lib;
    }
    machines.clear();
    pluginLibraries.clear();
}

pluginlib* player::getPluginlibByUri(const std::string &uri) {
    for (size_t i=0; i<pluginLibraries.size(); i++) {
		pluginlib *lib = pluginLibraries[i];
		if (lib->collection && lib->collection->get_uri() && (uri == lib->collection->get_uri())) {
			return lib;
		}
    }
	return 0;
}

void player::loadMachineLoaders() {
	// add input collection
	inputPluginCollection.setPlayer(this);
	pluginLibraries.push_back(new pluginlib("input",*this, &inputPluginCollection));
	// add recorder collection
	recorderPluginCollection.setPlayer(this);
	pluginLibraries.push_back(new pluginlib("recorder",*this, &recorderPluginCollection));

	// initialize rest like usual
	for (size_t i=0; i<pluginFolders.size(); i++) {
		initializeFolder(pluginFolders[i]);
	}
}

/*! \brief Returns the current state of the player
*/
player_state player::getPlayState() { 
	return playerState; 
}

/*! \brief Initializes internal player data.
	\return Returns false if player could not be initialized.

    initialize does the following:
		- initializes mixer
		- loads and indexes all dll's
		- generate built-in oscillator tables
		- initializes master and adds it to the stage
		- initializes a performance timer for CPU measurement
		- sets the player to stopped state
		- check if internal memory structures are aligned correctly

*/
bool player::initialize(int samplesPerSec) {

	this->masterInfo.samples_per_second=samplesPerSec;

    masterLoader=new master_pluginloader();

    master=createMachine(0, 0, "Master", masterLoader);
	if (!master) return false;

    master->initialize(0, 0, 0, 0, 0);
    master->tickAsync();

	wavePlayer.initialize(this);

	loadMachineLoaders();

	player::generateOscillatorTables();

	timer.start();
	workPos=0;

    resetMachines();
	playerState=player_state_stopped;
	
	driver.initialize(this);
    midiDriver.initialize(this);
	
	return true;
}

/*! \brief Returns the number of known machines.
*/
int player::getMachineLoaders() {
	return machines.size();
}

/*! \brief Returns a machine loader by index.
*/
pluginloader* player::getMachineLoader(int index) {
	int j=0;
	for (map<string, pluginloader*>::iterator i=machines.begin(); i!=machines.end(); ++i) {
		if (index==j) return i->second;
		j++;
	}
	return 0;
}

/*! \brief Returns a machine loader by name. If a loader contains more plugins '|' separates name  and pluginname
*/
pluginloader* player::getMachineLoader(std::string uri) {

    // master is hardcoded
    if (uri=="Master" || uri=="@zzub.org/master") return masterLoader;

	string keyName=uri;
	transform(keyName.begin(), keyName.end(), keyName.begin(), toLower);
	map<string, pluginloader*>::iterator i=machines.find(keyName);
	pluginloader* l=i==machines.end()?0:i->second;

	return l;
}

/*! \brief Lock the player. Restrict access on player resources to calling thread.

	This causes the player to halt all processing until unlock() is called.
	This method is used when you need to invoke non-thread safe methods on player 
	member objects, and return values from zzub::metaplugin and player methods.
*/
void player::lock() {
	playerLock.Lock();
}

/*! \brief Unlock the player.
*/
void player::unlock() {
	playerLock.Unlock();
}

/*!	\brief Lock sequencer and pattern data.

	lockTick() is used when clients need to only modify pattern and sequencer data.
	This allows the player thread to work the machines in the background.
*/
void player::lockTick() {
	playerTickLock.Lock();
}

/*!	\brief Unlock sequencer and pattern data.
*/
void player::unlockTick() {
	playerTickLock.Unlock();
}

/*! \brief Returns a machine loader by index.

*/
void player::resetMachines() {
	for (size_t i=0; i<getMachines(); i++) {
		zzub::metaplugin* machine=getMachine(i);
		if (!machine) continue;
		machine->resetMixer();
	}
}

/*!	\brief Tells the player to use another sequencer for playing.
	\param newseq If this is NULL, the player resumes playing from the built-in sequencer
*/
sequencer* player::setCurrentlyPlayingSequencer(sequencer* newseq) {
	sequencer* prev=currentlyPlayingSequencer;
	lock();
	if (!newseq)
		newseq=&song_sequencer;

	currentlyPlayingSequencer=newseq;
	//resetMachines();	// maybe not necessary, since we playerLock anyway?
	unlock();
	return prev;
}

/*	\brief Starts playing current song.
*/
void player::setPlayerState(player_state state) {
	if (playerState==state) return;

	lock();
	resetMachines();
	playerState=state;
	unlock();

	switch (playerState) {
		case player_state_stopped:
			for (size_t i=0; i<getMachines(); i++) {
				zzub::metaplugin* stopMachine=getMachine(i);
				if (stopMachine->getType()==zzub::plugin_type_generator)
					stopMachine->stop();	// (some?) effects seem to never start again (maybe because of softmuting which maybe should not be set?)
			}
			break;
	}
	
	// send event
	zzub_event_data eventData={event_type_player_state_changed};
	eventData.player_state_changed.player_state=(int)state;
	master->invokeEvent(eventData, false); // can also be fired if loop ends
}

/*	\brief Clears all data associated with current song from the player.
*/
void player::clear() {
	using namespace std;
	lock();
    setPlayerState(player_state_muted);
    recordParameters=false;
	unlock();

	song_sequencer.clear();

	waveTable.clear();

    midiInputMappings.clear();

	// delete all machines except master (which is always first)
	while (getMachines()>1) {
		zzub::metaplugin* m1=getMachine(1);
		deleteMachine(m1);
	}

	machineInstances.clear();
	machineInstances.push_back(master);

	master->clear();
	master->tick();

    setPlayerState(player_state_stopped);
	
	infoText = "";

}


/*	\brief Generates a new, unique name for a machine.
	\param The machine instance name you wish to generate a name for.
*/
std::string player::getNewMachineName(std::string machineUri) {
	using namespace std;
	
	pluginloader* loader=getMachineLoader(machineUri);
	string baseName;
	if (!loader) 
		baseName=machineUri; else
		baseName=loader->getInfo()->short_name;
	char pc[16];
	for (int i=0; i<9999; i++) {
		if (i==0) {
			strcpy(pc, baseName.substr(0,12).c_str());
		}  else
			sprintf(pc, "%s%i", baseName.substr(0,12).c_str(), i+1);
		zzub::metaplugin* m=getMachine(pc);
		if (!m) return pc;
	}
	return "_(error)"+baseName;
}

/*	\brief Sets or clears the solo machine.
	\param machine When machine is set to NULL, no machine is playing solo.
*/
void player::setSoloMachine(zzub::metaplugin* machine) {
	soloMachine=machine;
}

/*	\brief Returns current solo machine.
*/
zzub::metaplugin* player::getSoloMachine() {
	return soloMachine;
}

/*	\brief Returns the index of a machine.
*/
int player::getMachineIndex(zzub::metaplugin* machine) {
	for (int i=0; i<(int)machineInstances.size(); i++) {
		if (machine==getMachine(i)) {
			//machineLock.Unlock();
			return i;
		}
	}
	return -1;
}

/*	\brief Deletes a machine.

	Releases all connections, sequencer tracks, patterns and events associated with the machine.
*/
void player::deleteMachine(zzub::metaplugin* machine) {
	assert(machine!=master);

	lock();
	machineInstances.erase(machineInstances.begin()+getMachineIndex(machine));
    machine->initialized=false;

	for (size_t i=song_sequencer.getTracks(); i>0; i--) {
		if (song_sequencer.getTrack(i-1)->getMachine()==machine)
			song_sequencer.removeTrack(i-1);
	}

	unlock();
	
	machine->clear();

    // buzz has a gDeleteEvent declared in its machineinterface.h, but it is apparently never called
    // we tell out master
    zzub_event_data eventData={event_type_delete_plugin};
    eventData.delete_plugin.plugin=(zzub_plugin_t*)machine;
	master->invokeEvent(eventData, true);

    // flush pending events for this machine
    eventLock.Lock();
    for (deque<event_message>::iterator i = messageQueue.begin(); i!=messageQueue.end(); i++) {
        if (i->plugin == machine) {
            i = messageQueue.erase(i);
            --i;
        }
    }
    eventLock.Unlock();

	delete machine;
}

/*!	\brief Tests if the machine exists in the current song.

    Used internally to test for rogue pointers sent to CMICallbacks.
*/
bool player::machineExists(zzub::metaplugin* machine) {
	bool test=false;
	for (size_t i=0; i<machineInstances.size(); i++) {
		if (getMachine(i)==machine) {
			return true;
		}
	}

	return test;
}

/*!	\brief Returns number of machines in current song.
*/
size_t player::getMachines() { 
	size_t size=machineInstances.size(); 
	return size;
}

/*!	\brief Returns a machine by name.
*/
zzub::metaplugin* player::getMachine(std::string instanceName) {
	for (size_t i=0; i<machineInstances.size(); i++) {
		if (machineInstances[i]->getName()==instanceName) {
			zzub::metaplugin* m=(zzub::metaplugin*)machineInstances[i];
			return m;
		}
	}
	return 0;
}


/*!	\brief Returns a machine by index.
*/
zzub::metaplugin* player::getMachine(size_t index) {
	zzub::metaplugin* m=0;
	if (index>=0 && index<machineInstances.size()) 
		m=(zzub::metaplugin*)machineInstances[index];
	return m;
}


/*! \brief Returns information about the master.
*/
zzub::master_info* player::getMasterInfo() {
	return &masterInfo;
}

/*! \brief Returns currently playing position in the sequencer.
*/
int player::getSequencerPosition() {
	return song_sequencer.getPosition();
}

/*! \brief Sets the currently playing position in the sequencer.
*/
void player::setSequencerPosition(int v) {
	song_sequencer.setPosition(v);
}

void player::setSongBeginLoop(int v) {
	song_sequencer.beginOfLoop=v;
}

void player::setSongEndLoop(int v) {
	song_sequencer.endOfLoop=v;
}

void player::setSongBegin(int v) {
	song_sequencer.startOfSong=v;
}

int player::getSongBeginLoop() {
	return song_sequencer.beginOfLoop;
}

int player::getSongEndLoop() {
	return song_sequencer.endOfLoop;
}

int player::getSongBegin() {
	return song_sequencer.startOfSong;
}

void player::setLoopEnabled(bool enable) {
	song_sequencer.loop = enable;
}

bool player::getLoopEnabled() {
	return song_sequencer.loop;
}

void player::setSongEnd(int v) {
	song_sequencer.endOfSong=v;
}

int player::getSongEnd() {
	return song_sequencer.endOfSong;
}

/*! \brief Changes positions of sequencer tracks
*/
bool player::moveSequenceTrack(int oldTrack, int newTrack) {
	return song_sequencer.moveTrack(oldTrack, newTrack);
}

/*! \brief Removes a track from the sequencer
*/
bool player::removeSequenceTrack(int track) {
	return song_sequencer.removeTrack(track);
}

/*! \brief Returns the number of waves in the wavetable. Is usually 0xc8.
*/
size_t player::getWaves() {
	return waveTable.waves.size();
}

/*! \brief Returns the wave in the specified wavetable location.
*/
wave_info_ex* player::getWave(size_t index) {
	return &waveTable.waves[index];
}

int player::getWaveIndex(wave_info_ex* entry) {
	for (size_t i=0; i<getWaves(); i++) {
		if (getWave(i)==entry) return i;
	}
	return -1;
}


/*! \brief Returns the built-in wave player instance.
*/
wave_player* player::getWavePlayer() {
	return &wavePlayer;
}

/*! \brief Returns audio driver mixing rate.
*/
int player::samplesPerTick() {
	return masterInfo.samples_per_tick;
}

std::string player::getLoadWarnings() {
	return loadWarning;
}

std::string player::getLoadErrors() {
	return loadError;
}

sequence* player::getSequenceTrack(size_t i) {
	return song_sequencer.getTrack(i);
}

size_t player::getSequenceTracks() {
	return song_sequencer.getTracks();
}

sequencer* player::createSequencerRange(int fromRow, int fromTrack, int toRow, int toTrack) {
	return song_sequencer.createRangeSequencer(fromRow, fromTrack, toRow, toTrack);
}

pluginloader* player::createDummyLoader(int type, std::string fullName, int attributeCount, int numGlobalParams, int numTrackParams, zzub::parameter* params) {
    return new dummy_loader(type, fullName, attributeCount, numGlobalParams, numTrackParams, params);
}

metaplugin* player::createMachine(char* inputData, int inputSize, std::string machineName, pluginloader *loader) {
	metaplugin* machine=new metaplugin(this, loader);
	
    // add plugin to internal collection so it can successfully call the host callbacks during initialization
	machineInstances.push_back(machine);

    if (!machine->create(inputData, inputSize)) {
        // couldnt create, so remove from internal collection and clean up
        machineInstances.erase(machineInstances.begin() + machineInstances.size()-1);
        delete machine;
        return 0;
    }

    // automatically add NO_OUTPUT-machines to the master
    if (machine->isNoOutput())
	    getMaster()->addInput((metaplugin*)machine, 0x4000, 0x4000);

	machine->setName(machineName);

    return machine;
}

bool player::isBlackListed(std::string name) {
	transform(name.begin(), name.end(), name.begin(), toLower);
	for (size_t i=0; i<blacklist.size(); i++) {
		string keyName=blacklist[i];
		transform(keyName.begin(), keyName.end(), keyName.begin(), toLower);
		if (keyName==name) return true;
	}
	return false;
}


void player::updateTick(int workSamples) {
	player_state state=getPlayState();

	if (masterInfo.tick_position==0 && (state!=player_state_released && state!=player_state_muted)) {
		lockTick();
		lastTickPos=getWorkPosition();

		// the player machines process their own sequences

		for (size_t i=0; i<getMachines(); i++) {
			getMachine(i)->tick();
		}

		// this is for advancing the entire seqiuencer, each machine advances each track itself in tickmachine()
		if (state==player_state_playing) {
			if (!currentlyPlayingSequencer->advanceTick())
				setPlayerState(player_state_stopped);
		}

		unlockTick();

        workFracs+=masterInfo.samples_per_tick_frac;
	}

	mixSamples=masterInfo.samples_per_tick+floor(workFracs)-masterInfo.tick_position;	// mix only up to closest tick

	nextPosInTick=0;

    if (mixSamples>buffer_size) {
		mixSamples=buffer_size;
		if (mixSamples>(unsigned int)workSamples)
			mixSamples=workSamples;
		nextPosInTick=masterInfo.tick_position+mixSamples;	// remember where we are
	} else
	if (mixSamples>(unsigned int)workSamples) {
		mixSamples=workSamples;
		nextPosInTick=masterInfo.tick_position+mixSamples;	// remember where we are
    } else {
        double n;
        workFracs=modf(workFracs, &n);
	    nextPosInTick=0;
    }
}

void player::finishWork() {
	player_state state=getPlayState();

	metaplugin* masterInternal=(metaplugin*)master;
	if (state!=player_state_released)
		wavePlayer.work(masterInternal->machineBuffer, mixSamples, true);

	masterInfo.tick_position=nextPosInTick;
	workPos+=mixSamples;
}

bool player::workConnection(connection* conn, int numSamples) {
	metaplugin* outputMachine=(metaplugin*)conn->plugin_out;
	metaplugin* inputMachine=(metaplugin*)conn->plugin_in;
	float inAmp=conn->amp/(float)0x4000;
	float inPan=conn->pan/(float)0x4000;	// pan=0..2, 1=center
	bool status=false;

	//assert(conn->lastFrame!=outputMachine->lastWorkPos); // lr: created problems when loading tracks
	conn->lastFrame=outputMachine->lastWorkPos;

	// test if we already worked this machine (through a different effect chain)
	if (inputMachine->lastWorkPos==masterInfo.tick_position) {
		status=inputMachine->lastWorkState;
	} else {
		status=workMachine(inputMachine, numSamples);
	}

	// test for solo-state or if we're somehow muted and not supposed to play
	if (!inputMachine->isSoloMutePlaying()) {
		status=false;
	}

	if (outputMachine->doesInputMixing() && !outputMachine->isBypassed()) {
		if (status)
			outputMachine->input(inputMachine->machineBuffer, numSamples, inAmp); else
			outputMachine->input(0, 0, 0);
	}

	if (status && (!outputMachine->doesInputMixing() || outputMachine->isBypassed()) ) {
		AddS2SPanMC(outputMachine->machineBuffer, inputMachine->machineBuffer, numSamples, inAmp, inPan);
	}

	return status;
}


bool player::workMachine(metaplugin* machine, int numSamples) {
	assert(machine->lastWorkPos!=workPos);

	size_t inputConnections=machine->inConnections.size();
	bool inputState=inputConnections?false:true;

	memset(machine->machineBuffer[0], 0, numSamples*sizeof(float));
	memset(machine->machineBuffer[1], 0, numSamples*sizeof(float));

	int maxInputAmp=0;
	for (size_t i=0; i<inputConnections; i++) {
		if (machine->inConnections[i]->amp>maxInputAmp)
			maxInputAmp=machine->inConnections[i]->amp;
		if (workConnection(machine->inConnections[i], numSamples)) {
			inputState=true;
		}
	}

	// bestem flagg: 
	//	- hvis minst en av inn-maskinene var suksess s blir det WM_READWRITE p fx
	// TODO: m sette WM_NOIO nr maskinen ikke er koblet p noen, men alle pathene til dette stedet kommer via connections, s det er noe som mangler
	int flags;
	if (inputState) {
		// testing for doesInputMixing || !isBufferZero() fixes F:\audio\bmx\div\nool+ladproject - the longest travel.bmx
		// testing for maxInputAmp fixes F:\audio\bmx\orange.bmx
		if (inputConnections>0 && maxInputAmp>0 && (machine->doesInputMixing() || !isBufferZero(machine->machineBuffer, numSamples)))
//		if (inputConnections>0 && maxInputAmp>0)
			flags=zzub::process_mode_read_write; else
			flags=zzub::process_mode_write;
	} else {
		flags=zzub::process_mode_write;
	}

    // this if-test worked fine before the -1 - 1 transition, but now it cant be here anymore (clears the work buffer, but inside there is mode_write and ret=false all the way)
    // -> machineBuffer is all zeros, so this clears the mixBuffer, which is a good idea
    //if (((flags&zzub::process_mode_read)!=0) && !machine->doesInputMixing())
	memcpy(mixBuffer[0], machine->machineBuffer[0], numSamples*sizeof(float));
	memcpy(mixBuffer[1], machine->machineBuffer[1], numSamples*sizeof(float));

	bool ret=machine->work(mixBuffer, numSamples, flags);

	return ret;
}

void player::workStereo(float** buffer, int numSamples) {
	using namespace std;
	lock();
	
	SETABRPUN();
	double tempTime=timer.frame();
	int samplecount = numSamples;

    long bufpos = 0;
    while (numSamples > 0) {
        if (driver.isRecording()) {
            inputBuffer[0] = &buffer[0][bufpos];
            inputBuffer[1] = &buffer[1][bufpos];
        } else {
            inputBuffer[0] = inputBuffer[1] = 0;
        }

	    updateTick(numSamples);

	    player_state state = getPlayState();

	    memset(master->machineBuffer[0], 0, MAXBUFFERSAMPLES*sizeof(float));
	    memset(master->machineBuffer[1], 0, MAXBUFFERSAMPLES*sizeof(float));

	    if (state != player_state_muted && state != player_state_released)
		    workMachine(master, mixSamples);

	    finishWork();
        memcpy(&buffer[0][bufpos], master->machineBuffer[0], mixSamples*sizeof(float));
        memcpy(&buffer[1][bufpos], master->machineBuffer[1], mixSamples*sizeof(float));
	    numSamples -= mixSamples;
        bufpos += mixSamples;
    }
	
	workTime=timer.frame()-tempTime;
	// how long we needed / how much we could have needed
	double load=(100.0 * workTime * double(masterInfo.samples_per_second)) / double(samplecount);
	// slowly approach to new value
	cpuLoad += 0.1 * (load - cpuLoad);
	//SETGRADUN(); // don't turn it off

    midiDriver.poll();
	unlock();
}

std::string player::getBuzzUri(std::string name) {
    transform(name.begin(), name.end(), name.begin(), toLower);
    for (map<string, string>::iterator i=aliases.begin(); i!=aliases.end(); ++i) {
        string alias=i->first;
        transform(alias.begin(), alias.end(), alias.begin(), toLower);
        if (alias==name) return i->second;
    }
    return "";
}

std::string player::getBuzzName(std::string uri) {
    transform(uri.begin(), uri.end(), uri.begin(), toLower);
    for (map<string, string>::iterator i=aliases.begin(); i!=aliases.end(); ++i) {
        string id=i->second;
        transform(id.begin(), id.end(), id.begin(), toLower);
        if (id==uri) return i->first;
    }
    return "";
}

void player::midiEvent(unsigned short status, unsigned char data1, unsigned char data2) {
    // look up mapping(s) and send value to plugin
    int channel=status&0xF;
    for (size_t i=0; i<midiInputMappings.size(); i++) {
        midimapping& mm=midiInputMappings[i];
        if (mm.channel==channel && mm.controller==data1) {
            metaplugin* plugin=mm.machine;
            const parameter* param=plugin->getMachineParameter(mm.group, mm.track, mm.column);
            int minValue=param->value_min;
		    int maxValue=param->value_max;
            float delta=(float)((maxValue-minValue)) / 127.0f;

            plugin->setParameter(mm.group, mm.track, mm.column, minValue + (float)data2 * delta, true);
		    plugin->machine->process_events();
        }
    }
	
	// also send note events to plugins directly
	int command = (status & 0xf0) >> 4;
	if ((command == 8) || (command == 9)) {
		int velocity = (int)data2;
		if (command == 8)
			velocity = 0;
		for (size_t i=0; i<machineInstances.size(); i++) {
			zzub::metaplugin* m=(zzub::metaplugin*)machineInstances[i];
			if (m->machine)
				m->machine->midi_note(channel, (int)data1, velocity);
		}
	}
	else if (command == 0xb) {
		for (size_t i=0; i<machineInstances.size(); i++) {
			zzub::metaplugin* m=(zzub::metaplugin*)machineInstances[i];
			if (m->machine)
				m->machine->midi_control_change((int)data1, channel, (int)data2);
		}
	}

    // plus all midi messages should be sent as master-events, so ui's can pick these up
    zzub_event_data eventData={event_type_midi_control};
    eventData.midi_message.status=status;
    eventData.midi_message.data1=data1;
    eventData.midi_message.data2=data2;

    master->invokeEvent(eventData);
}

midimapping* player::addMidiMapping(metaplugin* plugin, size_t group, size_t track, size_t param, size_t channel, size_t controller) {
    midimapping mm={plugin, group, track, param, channel, controller };
    midiInputMappings.push_back(mm);
    return &midiInputMappings.back();
}

bool player::removeMidiMapping(metaplugin* plugin, size_t group, size_t track, size_t param) {
    for (size_t i=0; i<midiInputMappings.size(); i++) {
        midimapping& mm=midiInputMappings[i];
        if (mm.group==group && mm.track==track && mm.column==param) {
            midiInputMappings.erase(midiInputMappings.begin()+i);
            i--;
        }
    }
    return false;
}

midimapping* player::getMidiMapping(size_t index) {
    return &midiInputMappings[index];
}

size_t player::getMidiMappings() {
    return midiInputMappings.size();
}


void player::handleMessages() {
    while (messageQueue.size()) {
        event_message& message=messageQueue.front();
        message.event->invoke(message.data);
        eventLock.Lock();
        messageQueue.pop_front();
        eventLock.Unlock();
    }
}

float player::getBeatsPerMinute() {
	return float(master->getParameter(1, 0, 1));
}

int player::getTicksPerBeat() {
	return master->getParameter(1, 0, 2);
}

void player::setBeatsPerMinute(float bpm) {
	master->setParameter(1, 0, 1, (int)(bpm+0.5), false);
}

void player::setTicksPerBeat(int tpb) {
	master->setParameter(1, 0, 2, tpb, false);
}

} // namespace zzub
