/*****************************************************************************

        Biquad.cpp
        Copyright (c) 2002-2006 Laurent de Soras

--- Legal stuff ---

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*Tab=3***********************************************************************/



#if defined (_MSC_VER)
	#pragma warning (1 : 4130) // "'operator' : logical operation on address of string constant"
	#pragma warning (1 : 4223) // "nonstandard extension used : non-lvalue array converted to pointer"
	#pragma warning (1 : 4705) // "statement has no effect"
	#pragma warning (1 : 4706) // "assignment within conditional expression"
	#pragma warning (4 : 4786) // "identifier was truncated to '255' characters in the debug information"
	#pragma warning (4 : 4800) // "forcing value to bool 'true' or 'false' (performance warning)"
	#pragma warning (4 : 4355) // "'this' : used in base member initializer list"
#endif



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	"../basic/basic_def.h"
#include	"../basic/basic_fnc.h"
#include	"Biquad.h"
#include	"dsp_def.h"

#include	<cassert>
#include	<cmath>



namespace dsp
{



/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*
==============================================================================
Name: ctor
Description:
	Builds the filter, ready to work (transfer function is 1)
Throws: Nothing
==============================================================================
*/

Biquad::Biquad ()
{
	_z_eq_b [0] = 1;
	_z_eq_b [1] = 0;
	_z_eq_b [2] = 0;
	_z_eq_a [0] = 1;
	_z_eq_a [1] = 0;
	_z_eq_a [2] = 0;

	clear_buffers ();
}



/*
==============================================================================
Name: process_block
Description:
	Filters a block of samples. Can work in-place.
Input parameters:
	- src_ptr: Input sample block
	- nbr_spl: Number of samples to process. >= 0
Output parameters:
	- dest_ptr: Output sample block
Throws: Nothing
==============================================================================
*/

void	Biquad::process_block (float *dest_ptr, const float *src_ptr, long nbr_spl)
{
	assert (nbr_spl >= 0);

	if (nbr_spl == 0)
	{
		return;
	}

	// If we're not on a pair boudary, we process a single sample.
	if (_mem_pos != 0)
	{
		*dest_ptr = process_sample (*src_ptr);
		++ src_ptr;
		++ dest_ptr;
		-- nbr_spl;
	}

	if (nbr_spl == 0)
	{
		return;
	}

	long				half_nbr_spl = nbr_spl >> 1;
	long				index = 0;
	if (half_nbr_spl > 0)
	{
		float				mem_x [2];
		float				mem_y [2];
		mem_x [0] = _mem_x [0];
		mem_x [1] = _mem_x [1];
		mem_y [0] = _mem_y [0];
		mem_y [1] = _mem_y [1];

		do
		{

			float				x = src_ptr [index];
			mem_y [1] =   _z_eq_b [0] * x
			            + (  _z_eq_b [1] * mem_x [0]
			               + _z_eq_b [2] * mem_x [1])
			            - (  _z_eq_a [1] * mem_y [0]
			               + _z_eq_a [2] * mem_y [1]);

			mem_x [1] = x;
			dest_ptr [index] = mem_y [1];

			x = src_ptr [index + 1];
			mem_y [0] =   _z_eq_b [0] * x
			            + (  _z_eq_b [1] * mem_x [1]
			               + _z_eq_b [2] * mem_x [0])
			            - (  _z_eq_a [1] * mem_y [1]
			               + _z_eq_a [2] * mem_y [0]);

			mem_x [0] = x;
			dest_ptr [index + 1] = mem_y [0];
			index += 2;

			-- half_nbr_spl;
		}
		while (half_nbr_spl > 0);

		_mem_x [0] = mem_x [0];
		_mem_x [1] = mem_x [1];
		_mem_y [0] = mem_y [0];
		_mem_y [1] = mem_y [1];
	}

	// If number of samples was odd, there is one more to process.
	if ((nbr_spl & 1) > 0)
	{
		dest_ptr [index] = process_sample (src_ptr [index]);
	}
}


/*
==============================================================================
Name: process_block
Description:
	Filters a block of data, in-place
Input parameters:
	- nbr_spl: Number of samples to process, >= 0
Input/output parameters:
	- spl_ptr: Block of data to be filtered
Throws: Nothing
==============================================================================
*/

void	Biquad::process_block (float *spl_ptr, long nbr_spl)
{
	process_block (spl_ptr, spl_ptr, nbr_spl);
}



/*
==============================================================================
Name: clear_buffers
Description:
	Clears all filter buffers.
Throws: Nothing
==============================================================================
*/

void	Biquad::clear_buffers ()
{
	_mem_x [0] = 0;
	_mem_x [1] = 0;
	_mem_y [0] = 0;
	_mem_y [1] = 0;
	_mem_pos = 0;
}



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



}	// namespace dsp



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
